/********************************************************
 *
 *  Copyright (c) 2017 Sony Corporation
 *
 ********************************************************/

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/cdev.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/reboot.h>
#include <linux/vmalloc.h>
#include <linux/ioctl.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/interrupt.h>

#include <linux/errno.h>

#include <linux/of_platform.h>

//#define KDEBUG
#include "exttype.h"
#include "general_macro.h"

#include "elbdriver.h"

MODULE_LICENSE("GPL");

typedef struct {
    void*            addr;
    const uint       offset;
    const uint       size;
} LbmBarInfo_t;

//#define VSW_MAP_ENABLE 1
static LbmBarInfo_t elb_bar[] = {
#ifdef VSW_MAP_ENABLE
    { NULL, LBM_CS2_ADDRESS       , LBM_CS2_SIZE       },
    { NULL, LBM_CS2_CS_ADDRESS    , LBM_CS2_CS_SIZE    },
    { NULL, LBM_CS2_CADEC_ADDRESS , LBM_CS2_CADEC_SIZE },
    { NULL, LBM_CS2_VSW_ADDRESS   , LBM_CS2_VSW_SIZE   },
    { NULL, LBM_CS3_ADDRESS       , LBM_CS3_SIZE       },
#else
    { NULL, LBM_ADDR_CS2_CADEC       , LBM_SIZE_CS2_CADEC        },
    { NULL, LBM_ADDR_CS3_FPGA        , LBM_SIZE_CS3_FPGA         },
    { NULL, LBM_ADDR_CS3_CADEC_CONFIG, LBM_SIZE_CS3_CADEC_CONFIG },
#endif
};

#define NUM_OF_BAR (ARRAY_LENGTH(elb_bar))

DECLARE_WAIT_QUEUE_HEAD(vd_wait_queue);
static irqreturn_t elb_intr(int irq, void* data)
{
    wake_up_interruptible(&vd_wait_queue);

    return IRQ_HANDLED;
}

#ifdef VSW_MAP_ENABLE
static inline ushort get_ref_counter(uchar* base_addr)
{
    return UC2_TO_USHORT(ioread8(base_addr + 0x8a), ioread8(base_addr + 0x8b));
}

static inline uchar get_oe(uchar* base_addr)
{
    return ioread8(base_addr + 0x13);
}

static inline uchar* get_base_addr(void)
{
    return elb_bar[LBM_AREA_CS3].addr;
}

static inline void reset_handle(uchar* base_addr, ushort counter)
{
    if (counter != 0xffff) {
        return;
    }

    iowrite8(1, base_addr + 0x87);

    udelay(1*100);

    iowrite8(0, base_addr + 0x87);
}
#else
static inline ushort get_ref_counter(uchar* base_addr)
{
    return ioread16(base_addr + 0xa0);
}

static inline uchar get_oe(uchar* base_addr)
{
    return ioread16(base_addr + 0x24);
}

static inline uchar* get_base_addr(void)
{
    return elb_bar[LBM_BAR_CADEC].addr;
}

static inline void reset_handle(uchar* base_addr, ushort counter)
{
    if (counter != 0xffff) {
        return;
    }

    iowrite16(1, base_addr + 0x9c);

    udelay(1*100);

    iowrite16(0, base_addr + 0x9c);
}
#endif

static int elb_wait_vd(void)
{
#define VD_TIMEOUT (2*1000)
    int ret_val = interruptible_sleep_on_timeout(&vd_wait_queue, VD_TIMEOUT);

    static ushort pre_vcounter = 0;
    static uchar  pre_oe = 0;

    uchar* addr     = get_base_addr();
    ushort vcounter = get_ref_counter(addr);
    uchar oe        = get_oe(addr);

#if 0 //for debug
    if (vcounter != 0) {
        if ((oe == pre_oe) || (vcounter != (pre_vcounter+1))) {
            printk("%s %d: %04x(%04x), %2x (%2x)\n", __func__, __LINE__,
                                                     vcounter, pre_vcounter, pre_oe, oe);
        }
    }
#endif
    reset_handle(addr, vcounter);

    pre_vcounter = vcounter;
    pre_oe       = oe;

    if (ret_val < 0) {
        printk("%s: fault (%d). \n", __func__, ret_val);
        return ret_val;
    }

    return OK;
}

static int elb_ioctl_wait_vd(void)
{
    return elb_wait_vd();
}

static int lbm_ioctl_read16(void* base_addr, LbmdIoctl_t* ioctl_data)
{
    ushort* addr = base_addr + ioctl_data->offset;

    int     i;
    int     ret_val = 0;
    ushort  data;
    ushort* ptr = ioctl_data->buf;

    for (i = 0; i < ioctl_data->size / sizeof(data); ++i) {
        data = ioread16be(addr);
        ++ addr;
        ret_val = copy_to_user(&ptr[i], &data, sizeof(data));

        if (ret_val) {
            printk("Error : %s : copy_to_user failed [%d]\n", __func__, ret_val);
        }
    }

    return ret_val;
}

static int lbm_ioctl_read32(void* base_addr, LbmdIoctl_t* ioctl_data)
{
    uint* addr = base_addr + ioctl_data->offset;

    int   i;
    int   ret_val = 0;
    uint  data;
    uint* ptr = ioctl_data->buf;

    for (i = 0; i < ioctl_data->size / sizeof(data); ++i) {
        data = ioread32be(addr);
        ++ addr;
        ret_val = copy_to_user(&ptr[i], &data, sizeof(data));

        if (ret_val) {
            printk("Error : %s : copy_to_user failed [%d]\n", __func__, ret_val);
        }
    }

    return ret_val;
}

static int lbm_ioctl_write16(void* base_addr, LbmdIoctl_t* ioctl_data)
{
    ushort* addr = base_addr + ioctl_data->offset;

    int     i;
    int     ret_val = 0;
    ushort  data;
    ushort* ptr = ioctl_data->buf;

    for (i = 0; i < ioctl_data->size / sizeof(data); ++i) {
        ret_val = copy_from_user(&data, &ptr[i], sizeof(data));
        iowrite16be(data, addr);
        ++addr;

        if (ret_val) {
            printk("Error : %s : copy_from_user failed [%d]\n", __func__, ret_val);
        }
    }

    return ret_val;
}

static int lbm_ioctl_write32(void* base_addr, LbmdIoctl_t* ioctl_data)
{
    uint* addr = base_addr + ioctl_data->offset;

    int   i;
    int   ret_val = 0;
    uint  data;
    uint* ptr = ioctl_data->buf;

    for (i = 0; i < ioctl_data->size / sizeof(data); ++i) {
        ret_val = copy_from_user(&data, &ptr[i], sizeof(data));
        iowrite32be(data, addr);
        ++addr;

        if (ret_val) {
            printk("Error : %s : copy_from_user failed [%d]\n", __func__, ret_val);
        }
    }

    return ret_val;
}

long elb_fops_ioctl(struct file *filep, uint cmd, ulong arg)
{
    int             ret = OK;
    LbmdIoctl_t*    ioctl_data = (LbmdIoctl_t __user *)arg;
    void*           base_addr = NULL;

    if (NUM_OF_BAR <= ioctl_data->area) {
        printk("%s: Unknown area ordered. area = %lu\n", __func__, ioctl_data->area);
        return -EFAULT;
    }

    base_addr = elb_bar[ioctl_data->area].addr;

    switch (cmd) {
    case IOCTL_READ16:
        if (!access_ok(VERIFY_WRITE, (LbmdIoctl_t __user *)arg, _IOC_SIZE(cmd))) {
            kdebug("%s %s %d\n", __FILE__, __FUNCTION__, __LINE__);
            return -EFAULT;
        }

        ret = lbm_ioctl_read16(base_addr, ioctl_data);
        break;

    case IOCTL_WRITE16:
        if (!access_ok(VERIFY_READ, (LbmdIoctl_t __user *)arg, _IOC_SIZE(cmd))) {
            kdebug("%s %s %d\n", __FILE__, __FUNCTION__, __LINE__);
            return -EFAULT;
        }

        ret = lbm_ioctl_write16(base_addr, ioctl_data);
        break;

    case IOCTL_READ32:
        if (!access_ok(VERIFY_WRITE, (LbmdIoctl_t __user *)arg, _IOC_SIZE(cmd))) {
            kdebug("%s %s %d\n", __FILE__, __FUNCTION__, __LINE__);
            return -EFAULT;
        }

        ret = lbm_ioctl_read32(base_addr, ioctl_data);
        break;

    case IOCTL_WRITE32:
        if (!access_ok(VERIFY_READ, (LbmdIoctl_t __user *)arg, _IOC_SIZE(cmd))) {
            kdebug("%s %s %d\n", __FILE__, __FUNCTION__, __LINE__);
            return -EFAULT;
        }

        ret = lbm_ioctl_write32(base_addr, ioctl_data);
        break;

    case IOCTL_WAIT_VD:
        ret = elb_ioctl_wait_vd();
        break;

    default:
        printk("%s: Command should be read or write.\n", __func__);
        return ERROR;
    }

    return ret;
}

static int elb_fops_open(struct inode* inode, struct file* file)
{
    inode->i_private = inode;
    file->private_data = file;

    return OK;
}

static int elb_fops_close(struct inode* inode, struct file* file)
{
    return OK;
}

struct file_operations elb_fops = {
    .owner          = THIS_MODULE,
    .open           = elb_fops_open,
    .release        = elb_fops_close,
    .unlocked_ioctl = elb_fops_ioctl,
};

//------------------------------------------------------------------
static int          elb_devs  = 1;
static int          elb_major = 0;
static struct cdev  elb_cdev;
static int elb_io_init(void)
{
    dev_t dev = MKDEV(elb_major, 0);
    int ret_val  = 0;
    int major;
    int i;

    ret_val = alloc_chrdev_region(&dev, 0, elb_devs, DEVICE_NAME);
    if (ret_val) {
        return ret_val;
    }

    cdev_init(&elb_cdev, &elb_fops);
    elb_cdev.owner = THIS_MODULE;
    elb_major = major = MAJOR(dev);

    ret_val = cdev_add(&elb_cdev, MKDEV(elb_major, 0), elb_devs);
    if (ret_val) {
        return ret_val;
    }

    for (i = 0; i < NUM_OF_BAR; ++i) {
        elb_bar[i].addr = ioremap_nocache(elb_bar[i].offset, elb_bar[i].size);

        if (elb_bar[i].addr == NULL) {
            printk("%s %d: ERROR: Allocate addr=0x%08x, size_0x%08x\n",
                    __func__, __LINE__, elb_bar[i].offset, elb_bar[i].size);
            return -ENOMEM;
        }
    }

    return OK;
}

#define INTR_EXTERNAL_IRQ0 1
static int elb_intr_init(struct platform_device* pdev)
{
    int ret_val = OK;

    int virq = platform_get_irq(pdev, INTR_EXTERNAL_IRQ0);
    if (virq == NO_IRQ) {
        printk("%s %d: ERROR\n", __func__, __LINE__);
        return -EINVAL;
    }
    printk("virq : %d(%d)\n", virq, INTR_EXTERNAL_IRQ0);

    ret_val = request_irq(virq,
                          &elb_intr,
                          IRQF_TRIGGER_FALLING,
                          DEVICE_NAME,
                          NULL);
    if (ret_val < 0) {
        printk("%s: Failed to set interrupt handler(%d) %d\n", __func__, virq, ret_val);
        printk(" - code %d , EIO %d , EINVAL %d\n", ret_val, EIO, EINVAL);
    }

    return ret_val;
}

static int elb_io_exit(void)
{
    dev_t dev = MKDEV(elb_major, 0);
    int i;

    cdev_del(&elb_cdev);
    unregister_chrdev_region(dev, elb_devs);

    for (i = 0; i < NUM_OF_BAR; ++i) {
        if (elb_bar[i].addr != NULL) {
            iounmap(elb_bar[i].addr);
        }
    }

    return OK;
}

static int elb_probe(struct platform_device* pdev)
{
    int ret_val = ERROR;

    ret_val = elb_intr_init(pdev);
    if (ret_val < 0) {
        return ret_val;
    }

    ret_val = elb_io_init();

    printk("%s %s : %s driver installed. \n", __DATE__, __TIME__, DEVICE_NAME);

    return ret_val;
}

static int elb_remove(struct platform_device* pdev)
{
    int virq = 256;
    virq = platform_get_irq(pdev, INTR_EXTERNAL_IRQ0);
    if (virq == NO_IRQ) {
        printk("%s %d: ERROR\n", __func__, __LINE__);
        return -EINVAL;
    }
    printk("virq : %d(%d)\n", virq, INTR_EXTERNAL_IRQ0);

    free_irq(virq, NULL);

    elb_io_exit();

    printk("%s %s: driver %s unloaded\n", __DATE__, __TIME__, DEVICE_NAME);

    return OK;
}

static const struct of_device_id exintr_match[] = {
    { .compatible = "fsl,mpic-exintr", },
    { /* end of list */ },
};

static struct platform_driver exintr_driver = {
    .probe          = elb_probe,
    .remove         = elb_remove,
    .driver = {
        .name           = DEVICE_NAME,
        .owner          = THIS_MODULE,
        .of_match_table = exintr_match,
    },
};

module_platform_driver(exintr_driver);

MODULE_AUTHOR("Sony Corporation.");
MODULE_DESCRIPTION("Local Bus Memory driver");
